# Scrappit, Simple Reddit Scraper
# Copyright (C) 2024  Natan Junges <natanajunges@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass, field
from queue import Empty, PriorityQueue, Queue
from threading import Event, Thread
from time import sleep

from .common import ScrappitTask, ScrappitResult
from .api import RedditAPITask, RedditAPISubredditSort, RedditAPIT, RedditAPIUserWhere, RedditAPIUserSort, RedditAPICommentsSort, RedditAPI


@dataclass
class ScrappitDispatcherTask(ScrappitTask):
    priority: float = 0
    task_id: int = field(default=0, init=False, repr=False)


class ScrappitDispatcher(Thread):
    IDLE_SLEEP: float = 1 / 60

    def __init__(self) -> None:
        super().__init__()
        self.api: RedditAPI = RedditAPI()
        self.task_queue: PriorityQueue[ScrappitDispatcherTask] = PriorityQueue()
        self.task_id: int = 0
        self.result_queue: Queue[ScrappitResult] = Queue()
        self.running: Event = Event()

    def run(self) -> None:
        self.running.set()

        while self.running.is_set():
            if not self.task_queue.empty():
                task = self.task_queue.get()

                try:
                    json = getattr(self.api, task.task)(*task.args, **task.kwargs)
                    self.result_queue.put(ScrappitResult(task, json))
                except Exception as e:
                    self.result_queue.put(ScrappitResult(task, e))

                self.task_queue.task_done()
            else:
                sleep(self.IDLE_SLEEP)

    def stop(self) -> None:
        self.running.clear()

    def put_task(self, task: ScrappitDispatcherTask) -> ScrappitDispatcherTask:
        task.task_id = self.task_id
        self.task_id += 1
        self.task_queue.put(task)
        return task

    def get_result(self) -> ScrappitResult | None:
        try:
            result = self.result_queue.get_nowait()
            self.result_queue.task_done()
            return result
        except Empty:
            return None

    def get(self, endpoint: str, priority: float | None = None, **params: str) -> ScrappitDispatcherTask:
        if priority is None:
            priority = RedditAPITask.GET.value.priority

        return self.put_task(ScrappitDispatcherTask(RedditAPITask.GET.value.name, (endpoint,), params, priority))

    def listing(
        self, endpoint: str, before: str | None = None, after: str | None = None, priority: float | None = None, **params: str
    ) -> ScrappitDispatcherTask:
        if priority is None:
            priority = RedditAPITask.LISTING.value.priority

        return self.put_task(ScrappitDispatcherTask(RedditAPITask.LISTING.value.name, (endpoint, before, after), params, priority))

    def r_about(self, subreddit: str, priority: float | None = None) -> ScrappitDispatcherTask:
        if priority is None:
            priority = RedditAPITask.R_ABOUT.value.priority

        return self.put_task(ScrappitDispatcherTask(RedditAPITask.R_ABOUT.value.name, (subreddit,), priority=priority))

    def r(
        self,
        subreddit: str,
        sort: RedditAPISubredditSort = RedditAPISubredditSort.HOT,
        t: RedditAPIT = RedditAPIT.DAY,
        before: str | None = None,
        after: str | None = None,
        priority: float | None = None
    ) -> ScrappitDispatcherTask:
        if priority is None:
            priority = RedditAPITask.R.value.priority + sort.value.priority

            if sort in (RedditAPISubredditSort.TOP, RedditAPISubredditSort.CONTROVERSIAL):
                priority += t.value.priority
                priority /= 3
            else:
                priority /= 2

        return self.put_task(ScrappitDispatcherTask(RedditAPITask.R.value.name, (subreddit, sort, t, before, after), priority=priority))

    def user_about(self, username: str, priority: float | None = None) -> ScrappitDispatcherTask:
        if priority is None:
            priority = RedditAPITask.USER_ABOUT.value.priority

        return self.put_task(ScrappitDispatcherTask(RedditAPITask.USER_ABOUT.value.name, (username,), priority=priority))

    def user(
        self,
        username: str,
        where: RedditAPIUserWhere = RedditAPIUserWhere.OVERVIEW,
        sort: RedditAPIUserSort = RedditAPIUserSort.NEW,
        t: RedditAPIT = RedditAPIT.ALL,
        before: str | None = None,
        after: str | None = None,
        priority: float | None = None
    ) -> ScrappitDispatcherTask:
        if priority is None:
            priority = RedditAPITask.USER.value.priority + where.value.priority + sort.value.priority

            if sort in (RedditAPIUserSort.TOP, RedditAPIUserSort.CONTROVERSIAL):
                priority += t.value.priority
                priority /= 4
            else:
                priority /= 3

        return self.put_task(ScrappitDispatcherTask(RedditAPITask.USER.value.name, (username, where, sort, t, before, after), priority=priority))

    def comments(
        self, article: str, sort: RedditAPICommentsSort = RedditAPICommentsSort.CONFIDENCE, comment: str | None = None, priority: float | None = None
    ) -> ScrappitDispatcherTask:
        if priority is None:
            priority = RedditAPITask.COMMENTS.value.priority + sort.value.priority
            priority /= 2

        return self.put_task(ScrappitDispatcherTask(RedditAPITask.COMMENTS.value.name, (article, sort, comment), priority=priority))

    def api_morechildren(
        self, link_id: str, children: list[str], sort: RedditAPICommentsSort = RedditAPICommentsSort.CONFIDENCE, priority: float | None = None
    ) -> ScrappitDispatcherTask:
        if priority is None:
            priority = RedditAPITask.API_MORECHILDREN.value.priority + sort.value.priority
            priority /= 2

        return self.put_task(ScrappitDispatcherTask(RedditAPITask.API_MORECHILDREN.value.name, (link_id, children, sort), priority=priority))



# Scrappit, Simple Reddit Scraper
# Copyright (C) 2024  Natan Junges <natanajunges@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass, field
from enum import Enum
from http.cookiejar import DefaultCookiePolicy
from time import sleep, time
from typing import ClassVar

from fake_useragent import UserAgent
from requests import Session, Timeout
from requests.exceptions import RetryError

from .common import JSON


@dataclass
class RedditAPIItem:
    name: str
    priority: float


class RedditAPITask(Enum):
    GET = RedditAPIItem("get", 0 / 8)
    LISTING = RedditAPIItem("listing", 1 / 8)
    R_ABOUT = RedditAPIItem("r_about", 6 / 8)
    R = RedditAPIItem("r", 2 / 8)
    USER_ABOUT = RedditAPIItem("user_about", 7 / 8)
    USER = RedditAPIItem("user", 3 / 8)
    COMMENTS = RedditAPIItem("comments", 4 / 8)
    API_MORECHILDREN = RedditAPIItem("api_morechildren", 5 / 8)


class RedditAPISubredditSort(Enum):
    HOT = RedditAPIItem("hot", 2 / 5)
    NEW = RedditAPIItem("new", 0 / 5)
    TOP = RedditAPIItem("top", 4 / 5)
    CONTROVERSIAL = RedditAPIItem("controversial", 3 / 5)
    RISING = RedditAPIItem("rising", 1 / 5)


class RedditAPIT(Enum):
    HOUR = RedditAPIItem("hour", 0 / 6)
    DAY = RedditAPIItem("day", 1 / 6)
    WEEK = RedditAPIItem("week", 2 / 6)
    MONTH = RedditAPIItem("month", 3 / 6)
    YEAR = RedditAPIItem("year", 4 / 6)
    ALL = RedditAPIItem("all", 5 / 6)


class RedditAPIUserWhere(Enum):
    OVERVIEW = RedditAPIItem("overview", 0 / 3)
    SUBMITTED = RedditAPIItem("submitted", 2 / 3)
    COMMENTS = RedditAPIItem("comments", 1 / 3)


class RedditAPIUserSort(Enum):
    HOT = RedditAPIItem("hot", 1 / 4)
    NEW = RedditAPIItem("new", 0 / 4)
    TOP = RedditAPIItem("top", 3 / 4)
    CONTROVERSIAL = RedditAPIItem("controversial", 2 / 4)


class RedditAPICommentsSort(Enum):
    CONFIDENCE = RedditAPIItem("confidence", 1 / 6)
    TOP = RedditAPIItem("top", 4 / 6)
    NEW = RedditAPIItem("new", 0 / 6)
    CONTROVERSIAL = RedditAPIItem("controversial", 3 / 6)
    OLD = RedditAPIItem("old", 5 / 6)
    QA = RedditAPIItem("qa", 2 / 6)


@dataclass
class RedditAPI:
    BASE_URL: ClassVar[str] = "https://reddit.com"
    TIMEOUT: ClassVar[int] = 10
    MAX_TRIES: ClassVar[int] = 3

    session: Session = field(default_factory=Session, init=False, repr=False)
    user_agent: UserAgent = field(default_factory=UserAgent, init=False, repr=False)
    requests_remaining: bool = field(default=True, init=False, repr=False)
    reset_time: float = field(default=0, init=False, repr=False)

    def __post_init__(self) -> None:
        self.session.headers = {"User-Agent": self.user_agent.random}
        self.session.cookies.set_policy(DefaultCookiePolicy(allowed_domains=[]))

    def get(self, endpoint: str, **params: str) -> JSON:
        params["raw_json"] = "1"

        for _ in range(self.MAX_TRIES):
            now = time()

            if now > self.reset_time:
                self.requests_remaining = True
            elif not self.requests_remaining:
                sleep(self.reset_time - now)
                self.session.headers["User-Agent"] = self.user_agent.random
                self.requests_remaining = True

            try:
                response = self.session.get(f"{self.BASE_URL}{endpoint}.json", params=params, timeout=self.TIMEOUT)
            except Timeout:
                continue

            if response.status_code not in (200, 429):
                response.raise_for_status()

            now = time()
            self.requests_remaining = bool(float(response.headers["X-Ratelimit-Remaining"]))
            self.reset_time = now + int(response.headers["X-Ratelimit-Reset"])

            if response.status_code == 429:
                sleep(max(self.TIMEOUT, self.reset_time - now))
                self.session.headers["User-Agent"] = self.user_agent.random
                continue

            return response.json()

        raise RetryError()

    def listing(self, endpoint: str, before: str | None = None, after: str | None = None, **params: str) -> JSON:
        params["limit"] = "100"

        if before:
            params["before"] = before
        elif after:
            params["after"] = after

        return self.get(endpoint, **params)

    def r_about(self, subreddit: str) -> JSON:
        return self.get(f"/r/{subreddit}/about")

    def r(
        self,
        subreddit: str,
        sort: RedditAPISubredditSort = RedditAPISubredditSort.HOT,
        t: RedditAPIT = RedditAPIT.DAY,
        before: str | None = None,
        after: str | None = None
    ) -> JSON:
        endpoint = f"/r/{subreddit}/{sort.value.name}"

        if sort in (RedditAPISubredditSort.TOP, RedditAPISubredditSort.CONTROVERSIAL):
            return self.listing(endpoint, before, after, t=t.value.name)

        return self.listing(endpoint, before, after)

    def user_about(self, username: str) -> JSON:
        return self.get(f"/user/{username}/about")

    def user(
        self,
        username: str,
        where: RedditAPIUserWhere = RedditAPIUserWhere.OVERVIEW,
        sort: RedditAPIUserSort = RedditAPIUserSort.NEW,
        t: RedditAPIT = RedditAPIT.ALL,
        before: str | None = None,
        after: str | None = None
    ) -> JSON:
        endpoint = f"/user/{username}/{where.value.name}"

        if sort in (RedditAPIUserSort.TOP, RedditAPIUserSort.CONTROVERSIAL):
            return self.listing(endpoint, before, after, sort=sort.value.name, t=t.value.name)

        return self.listing(endpoint, before, after, sort=sort.value.name)

    def comments(self, article: str, sort: RedditAPICommentsSort = RedditAPICommentsSort.CONFIDENCE, comment: str | None = None) -> JSON:
        endpoint = f"/comments/{article}"

        if comment:
            return self.get(endpoint, sort=sort.value.name, comment=comment)

        return self.get(endpoint, sort=sort.value.name)

    def api_morechildren(self, link_id: str, children: list[str], sort: RedditAPICommentsSort = RedditAPICommentsSort.CONFIDENCE) -> JSON:
        return self.get("/api/morechildren", api_type="json", link_id=link_id, children=",".join(children), sort=sort.value.name)



# Scrappit, Simple Reddit Scraper
# Copyright (C) 2024  Natan Junges <natanajunges@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from queue import Empty, Queue
from threading import Thread

from .common import ScrappitTask, ScrappitResult
from .api import SubredditSort, SubredditT, UserWhere, UserSort, UserT, CommentsSort
from .dispatcher import ScrappitDispatcherTask, ScrappitDispatcher


class ScrappitScheduler(Thread):
    IDLE_SLEEP: float = 1 / 60

    def __init__(self) -> None:
        super().__init__()
        self.dispatcher = ScrappitDispatcher()
        self.task_queue: Queue[ScrappitTask] = Queue()
        self.result_queue: Queue[ScrappitResult] = Queue()

    def run(self) -> None:
        self.dispatcher.start()
        self.dispatcher.running.wait()

        while self.dispatcher.running.is_set():
            pass

    def stop(self) -> None:
        self.dispatcher.stop()

    def put_task(self, task: ScrappitTask) -> ScrappitTask:
        self.task_queue.put(task)
        return task

    def get_result(self) -> ScrappitResult | None:
        try:
            result = self.result_queue.get_nowait()
            self.result_queue.task_done()
            return result
        except Empty:
            return None

    def get(self, endpoint: str, **params: str) -> ScrappitDispatcherTask:
        return self.dispatcher.get(0, endpoint, **params)

    def listing(self, endpoint: str, before: str | None = None, after: str | None = None, **params: str) -> ScrappitDispatcherTask:
        return self.dispatcher.listing(0, endpoint, before, after, **params)

    def listing_full(self, endpoint: str, limit: int = 0, **params: str) -> ScrappitTask:
        return self.put_task(ScrappitTask("listing_full", (endpoint, limit), params))

    def r_about(self, subreddit: str) -> ScrappitDispatcherTask:
        return self.dispatcher.r_about(0, subreddit)

    def r(
        self,
        subreddit: str,
        sort: SubredditSort = SubredditSort.HOT,
        t: SubredditT = SubredditT.DAY,
        before: str | None = None,
        after: str | None = None
    ) -> ScrappitDispatcherTask:
        return self.dispatcher.r(0, subreddit, sort, t, before, after)

    def r_full(self, subreddit: str, sort: SubredditSort = SubredditSort.HOT, t: SubredditT = SubredditT.DAY, limit: int = 0) -> ScrappitTask:
        return self.put_task(ScrappitTask("r_full", (subreddit, sort, t, limit)))

    def user_about(self, username: str) -> ScrappitDispatcherTask:
        return self.dispatcher.user_about(0, username)

    def user(
        self,
        username: str,
        where: UserWhere = UserWhere.OVERVIEW,
        sort: UserSort = UserSort.NEW,
        t: UserT = UserT.ALL,
        before: str | None = None,
        after: str | None = None
    ) -> ScrappitDispatcherTask:
        return self.dispatcher.user(0, username, where, sort, t, before, after)

    def user_full(
        self, username: str, where: UserWhere = UserWhere.OVERVIEW, sort: UserSort = UserSort.NEW, t: UserT = UserT.ALL, limit: int = 0
    ) -> ScrappitTask:
        return self.put_task(ScrappitTask("user_full", (username, where, sort, t, limit)))

    def comments(self, article: str, sort: CommentsSort = CommentsSort.CONFIDENCE, comment: str | None = None) -> ScrappitDispatcherTask:
        return self.dispatcher.comments(0, article, sort, comment)

    def api_morechildren(self, link_id: str, children: list[str], sort: CommentsSort = CommentsSort.CONFIDENCE) -> ScrappitDispatcherTask:
        return self.dispatcher.api_morechildren(0, link_id, children, sort)

    def comments_full(self, article: str, sort: CommentsSort = CommentsSort.CONFIDENCE, comment: str | None = None, limit: int = 0) -> ScrappitTask:
        return self.put_task(ScrappitTask("comments_full", (article, sort, comment, limit)))



# Scrappit, Simple Reddit Scraper
# Copyright (C) 2024  Natan Junges <natanajunges@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass, field
from typing import TypeAlias

JSON: TypeAlias = dict[str, "JSON"] | list["JSON"] | str | int | float | bool | None


@dataclass
class ScrappitTask:
    task: str = field(compare=False)
    args: tuple = field(compare=False)
    kwargs: dict = field(default_factory=dict, compare=False)


@dataclass
class ScrappitResult:
    task: ScrappitTask
    value: JSON | Exception
